/*
 *  ArchivedBitmapTranslator.cpp 1.0.0 (May 5th, 1998)
 *
 *  Translator add-on for reading and writing archived BBitmaps
 *  Written by Edmund Vermeulen <edmundv@xs4all.nl>
 *  Public domain
 */


// Includes
#include <DataIO.h>
#include <InterfaceDefs.h>
#include <TranslatorAddOn.h>
#include <TranslatorFormats.h>


// Defines
#define ARCHIVED_BITMAP_FORMAT 'arBB'


// Handler information
char translatorName[] = "ArchivedBitmapTranslator";
char translatorInfo[] = "By Edmund Vermeulen <edmundv@xs4all.nl>";
int32 translatorVersion = 100;

// Supported input formats
translation_format inputFormats[] =
{
	{ ARCHIVED_BITMAP_FORMAT, B_TRANSLATOR_BITMAP, 0.4, 0.9, "image/x-be-archived-bitmap", "Archived BBitmap" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.4, 0.6, "image/x-be-bitmap", "Be bitmap" },
	{ 0, 0, 0, 0, 0, 0 }
};

// Supported output formats
translation_format outputFormats[] =
{
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.4, 0.6, "image/x-be-bitmap", "Be bitmap" },
	{ ARCHIVED_BITMAP_FORMAT, B_TRANSLATOR_BITMAP, 0.4, 0.9, "image/x-be-archived-bitmap", "Archived BBitmap" },
	{ 0, 0,	0, 0, 0, 0 }
};


// Calculate number of bytes in one BBitmap row
static int32
GetRowBytes(int32 left, int32 right, color_space colors)
{
	int32 rowBytes = right - left + 1;
	switch (colors)
	{
		case B_MONOCHROME_1_BIT:
			rowBytes /= 8;
			break;

		case B_GRAYSCALE_8_BIT:
		case B_COLOR_8_BIT:
			break;

		case B_RGB_16_BIT:
			rowBytes *= 2;
			break;

		case B_RGB_32_BIT:
			rowBytes *= 4;
			break;

		default:  // Unknown
			rowBytes = 0;
			break;
	}

	// Long word align
	return ((rowBytes + 3) / 4) * 4;
}


// Read an archived BBitmap and convert it to B_TRANSLATOR_BITMAP
static status_t
ReadArchive(BPositionIO *inSource, BPositionIO *outDestination)
{
	// Unflatten the archive from the input stream
	BMessage archive;
	if (archive.Unflatten(inSource) != B_OK)
		return B_NO_TRANSLATOR;

	// Rehydrate the bitmap from the archive
	BBitmap arc_bitmap(&archive);
	if (!arc_bitmap.IsValid())
		return B_NO_TRANSLATOR;

	// Create B_TRANSLATOR_BITMAP header
	TranslatorBitmap hdr;
	hdr.magic = B_TRANSLATOR_BITMAP;
	hdr.bounds = arc_bitmap.Bounds();
	hdr.colors = arc_bitmap.ColorSpace();
	hdr.rowBytes = arc_bitmap.BytesPerRow();
	hdr.dataSize = arc_bitmap.BitsLength();

	// First convert the header to the correct endianess
	swap_data(B_UINT32_TYPE, &(hdr.magic), sizeof(uint32), B_SWAP_BENDIAN_TO_HOST);
	swap_data(B_RECT_TYPE, &(hdr.bounds), sizeof(BRect), B_SWAP_BENDIAN_TO_HOST);
	swap_data(B_UINT32_TYPE, &(hdr.rowBytes), sizeof(uint32), B_SWAP_BENDIAN_TO_HOST);
	swap_data(B_UINT32_TYPE, &(hdr.colors), sizeof(color_space), B_SWAP_BENDIAN_TO_HOST);
	swap_data(B_UINT32_TYPE, &(hdr.dataSize), sizeof(uint32), B_SWAP_BENDIAN_TO_HOST);

	// Write the header out
	if (outDestination->Write(&hdr, sizeof(TranslatorBitmap)) != sizeof(TranslatorBitmap))
		return B_ERROR;

	// Write all of the bitmap data in one go
	if (outDestination->Write(arc_bitmap.Bits(), arc_bitmap.BitsLength()) != arc_bitmap.BitsLength())
		return B_ERROR;

	return B_OK;
}


// Take a B_TRANSLATOR_BITMAP, and write it out as an archived BBitmap
static status_t
WriteArchive(BPositionIO *inSource, BPositionIO *outDestination)
{
	// Read in the header
	TranslatorBitmap hdr;
	if (inSource->Read(&hdr, sizeof(TranslatorBitmap) ) != sizeof(TranslatorBitmap))
		return false;

	// Convert the header to the correct endianess
	swap_data(B_UINT32_TYPE, &(hdr.magic), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data(B_RECT_TYPE, &(hdr.bounds), sizeof(BRect), B_SWAP_HOST_TO_BENDIAN);
	swap_data(B_UINT32_TYPE, &(hdr.rowBytes), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data(B_UINT32_TYPE, &(hdr.colors), sizeof(color_space), B_SWAP_HOST_TO_BENDIAN);
	swap_data(B_UINT32_TYPE, &(hdr.dataSize), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);

	// Create a BBitmap
	BBitmap bitmap(hdr.bounds, hdr.colors);
	if (!bitmap.IsValid())
		return B_ERROR;
	
	// Read the bitmap data into the BBitmap
	if (inSource->Read(bitmap.Bits(), bitmap.BitsLength()) != bitmap.BitsLength())
		return B_ERROR;

	// Archive the bitmap to the output stream
	BMessage archive;
	if (bitmap.Archive(&archive) != B_OK ||
	    archive.Flatten(outDestination) != B_OK)
	{
		return B_ERROR;
	}

	return B_OK;
}


// Identify function
status_t
Identify(
	BPositionIO *inSource,
	const translation_format *inFormat,
	BMessage * /*ioExtension*/,
	translator_info *outInfo,
	uint32 outType)
{
	if (outType && outType != B_TRANSLATOR_BITMAP && outType != ARCHIVED_BITMAP_FORMAT)
		return B_NO_TRANSLATOR;

	// Check if it is B_TRANSLATOR_BITMAP
	char data[32];
	if (inSource->Read(data, 32) != 32)
		return B_NO_TRANSLATOR;

	TranslatorBitmap *hdr = reinterpret_cast<TranslatorBitmap *>(data);

	// Convert the header to the correct endianess
	swap_data(B_UINT32_TYPE, &(hdr->magic), sizeof(uint32), B_SWAP_BENDIAN_TO_HOST);
	swap_data(B_RECT_TYPE, &(hdr->bounds), sizeof(BRect), B_SWAP_BENDIAN_TO_HOST);
	swap_data(B_UINT32_TYPE, &(hdr->rowBytes), sizeof(uint32), B_SWAP_BENDIAN_TO_HOST);
	swap_data(B_UINT32_TYPE, &(hdr->colors), sizeof(color_space), B_SWAP_BENDIAN_TO_HOST);
	swap_data(B_UINT32_TYPE, &(hdr->dataSize), sizeof(uint32), B_SWAP_BENDIAN_TO_HOST);

	if (hdr->magic == B_TRANSLATOR_BITMAP &&
	    hdr->bounds.IsValid() &&
	    hdr->rowBytes == GetRowBytes(hdr->bounds.left, hdr->bounds.right, hdr->colors) &&
	    hdr->dataSize == int32(hdr->bounds.bottom - hdr->bounds.top + 1) * hdr->rowBytes)
	{
		inFormat = &inputFormats[1];
	}
	else
	{
		// Check if it is an archived BBitmap
		inSource->Seek(0, SEEK_SET);
		BMessage archive;
		// It's rather inefficient to have to read in the whole file
		// just to identify it, but there you go...
		if (archive.Unflatten(inSource) == B_OK &&
		    validate_instantiation(&archive, "BBitmap"))
		{
			inFormat = &inputFormats[0];
		}
		else
		{
			return B_NO_TRANSLATOR;
		}
	}

	// Set the DATAInfo to the recognized type
	outInfo->type = inFormat->type;
	outInfo->group = inFormat->group;
	outInfo->quality = inFormat->quality;
	outInfo->capability = inFormat->capability;
	strcpy(outInfo->name, inFormat->name);
	strcpy(outInfo->MIME, inFormat->MIME);

	return B_OK;
}


// Translate function
status_t
Translate(
	BPositionIO *inSource,
	const translator_info *inInfo,
	BMessage * /*ioExtension*/,
	uint32 outFormat,
	BPositionIO *outDestination)
{
	// Default output format
	if (outFormat == 0)
		outFormat = B_TRANSLATOR_BITMAP;

	// Read an archive bitmap
	if (inInfo->type == ARCHIVED_BITMAP_FORMAT && outFormat == B_TRANSLATOR_BITMAP)
		return ReadArchive(inSource, outDestination);

	// Write an archived bitmap
	if (inInfo->type == B_TRANSLATOR_BITMAP && outFormat == ARCHIVED_BITMAP_FORMAT)
		return WriteArchive(inSource, outDestination);

	return B_NO_TRANSLATOR;
}
